# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018-2019 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


_clamp = lambda val: max(min(val, 255), 0)
_convert_input = lambda val: _clamp(int(val * 255.0)) if isinstance(val, float) else _clamp(int(val))


class Color(object):
  """
  Accepts integers in [0, 255] or floats in [0.0, 1.0] as input.
  """

  __slots__ = ['__a', '__b', '__g', '__r']

  def __init__(self, r, g, b, a=1.0):
    self.__r = _convert_input(r)
    self.__g = _convert_input(g)
    self.__b = _convert_input(b)
    self.__a = _convert_input(a)

  def __repr__(self):
    return 'Color(R:{0}, G:{1}, B:{2}, A:{3})'.format(self.r, self.g, self.b, self.a)

  def __int__(self):
    return ((self.__r & 0xff) << 24) | ((self.__g & 0xff) << 16) | ((self.__b & 0xff) << 8) | (self.__a & 0xff)

  @property
  def r(self):
    return self.__r

  @r.setter
  def r(self, value):
    self.__r = _convert_input(value)

  @property
  def g(self):
    return self.__g

  @g.setter
  def g(self, value):
    self.__g = _convert_input(value)

  @property
  def b(self):
    return self.__b

  @b.setter
  def b(self, value):
    self.__b = _convert_input(value)

  @property
  def a(self):
    return self.__a

  @a.setter
  def a(self, value):
    self.__a = _convert_input(value)

  @property
  def opaque(self):
      return self.__a == 255

  def alpha_blend(self, dst):
    """
    "This" instance refers to src (foreground)

    :param dst: destination color (background)
    :return: blended color
    """
    src_r = float(self.__r) / 255.0
    src_g = float(self.__g) / 255.0
    src_b = float(self.__b) / 255.0
    src_a = float(self.__a) / 255.0
    dst_r = float(dst.r) / 255.0
    dst_g = float(dst.g) / 255.0
    dst_b = float(dst.b) / 255.0
    dst_a = float(dst.a) / 255.0

    if dst.opaque:
      a = 1.0
      r = (src_r * src_a) + (dst_r * dst_a * (1 - src_a))
      g = (src_g * src_a) + (dst_g * dst_a * (1 - src_a))
      b = (src_b * src_a) + (dst_b * dst_a * (1 - src_a))
    else:
      a = src_a + dst_a * (1.0 - src_a)
      if a == 0.0:
        raise ValueError('both source and destination alphas were 0')
      r = ((src_r * src_a) + (dst_r * dst_a * (1 - src_a))) / a
      g = ((src_g * src_a) + (dst_g * dst_a * (1 - src_a))) / a
      b = ((src_b * src_a) + (dst_b * dst_a * (1 - src_a))) / a
    return Color(r, g, b, a)


def color_from_int(color):
  color= int(color)
  R = (color >> 24) & 0xff
  G = (color >> 16) & 0xff
  B = (color >> 8) & 0xff
  A = color & 0xff
  return Color(R, G, B, A)


def string_to_color(color):
  try:
    return color_from_int(color)
  except:
    o_color = color
    color = str(color).strip().lower()
    color = __str_colors.get(color, None)
    if color:
      return color
    raise ValueError("Cannot parse '{}' to color".format(o_color))


__str_colors = {
  'red' : Color(255, 0, 0),
  'green' : Color(0, 255, 0),
  'blue' : Color(0, 0, 255),
  'black' : Color(0, 0, 0),
  'white' : Color(255, 255, 255),
  'cyan' : Color(0, 255, 255),
  'magenta' : Color(255, 0, 255),
  'yellow' : Color(255, 255, 0),
  'transparent' : Color(0, 0, 0, 0)
}



import ctypes as _ctypes
import threading

from .ctypes_utils import cast, c_int_p, c_int32_p, c_uint32_p, c_int64_p, c_uint64_p, c_float_p, c_double_p


class TLS_Holder:
  __slots__ = ["_buffer", "_size",
              "_c_int_p", "_c_int32_p", "_c_uint32_p", "_c_int64_p", "_c_uint64_p", "_c_float_p", "_c_double_p"
  ]

  def _grow_arrays(self, size):
    if size > self._size:
      bfr = _ctypes.c_buffer(size)
      self._c_int_p = cast(bfr, c_int_p)
      self._c_int32_p = cast(bfr, c_int32_p)
      self._c_uint32_p = cast(bfr, c_uint32_p)
      self._c_int64_p = cast(bfr, c_int64_p)
      self._c_uint64_p = cast(bfr, c_uint64_p)
      self._c_float_p = cast(bfr, c_float_p)
      self._c_double_p = cast(bfr, c_double_p)
      self._buffer = bfr
      self._size = size

  def __init__(self):
    self._size = 0
    self._grow_arrays(1024)


class MessagesTLS(threading.local):
  def __init__(self):
    super(MessagesTLS, self).__init__()
    self._holder = TLS_Holder()


_tls = MessagesTLS()


int_buffer = lambda: _tls._holder._c_int_p
int32_buffer = lambda: _tls._holder._c_int32_p
uint32_buffer = lambda: _tls._holder._c_uint32_p
int64_buffer = lambda: _tls._holder._c_int64_p
uint64_buffer = lambda: _tls._holder._c_uint64_p
float_buffer = lambda: _tls._holder._c_float_p
double_buffer = lambda: _tls._holder._c_double_p


grow_buffer_if_needed_int = lambda size: _tls._holder._grow_arrays(size * 4)
grow_buffer_if_needed_long = lambda size: _tls._holder._grow_arrays(size * 8)



import ctypes


################################################################################
# Used to populate the robot model metadata structure
################################################################################


class _MetaDataAnonActuatorCTypes(ctypes.Structure):
  _fields_ = [("actuator_type_", ctypes.c_int)]


class _MetaDataAnonBracketCTypes(ctypes.Structure):
  _fields_ = [("bracket_type_", ctypes.c_int)]


class _MetaDataAnonJointCTypes(ctypes.Structure):
  _fields_ = [("joint_type_", ctypes.c_int)]


class _MetaDataAnonLinkCTypes(ctypes.Structure):
  _fields_ = [
    ("link_type_", ctypes.c_int),
    ("input_type_", ctypes.c_int),
    ("output_type_", ctypes.c_int),
    ("extension_", ctypes.c_float),
    ("twist_", ctypes.c_float)
  ]


class _MetaDataAnonEndEffectorCTypes(ctypes.Structure):
  _fields_ = [("end_effector_type_", ctypes.c_int)]


class _MetaDataUnionCTypes(ctypes.Union):
  _anonymous_ = ("actuator_type_struct_", "bracket_type_struct_", "joint_type_struct_", "link_type_struct_", "end_effector_type_struct_")
  _fields_ = [
    ("actuator_type_struct_", _MetaDataAnonActuatorCTypes),
    ("bracket_type_struct_", _MetaDataAnonBracketCTypes),
    ("joint_type_struct_", _MetaDataAnonJointCTypes),
    ("link_type_struct_", _MetaDataAnonLinkCTypes),
    ("end_effector_type_struct_", _MetaDataAnonEndEffectorCTypes)
  ]

################################################################################
# Anonymous unions
################################################################################


class _IoPinUnion(ctypes.Union):
  _fields_ = [
    ("int_value_", ctypes.c_int64),
    ("float_value_", ctypes.c_float)
  ]


################################################################################
# Public facing
################################################################################


class HebiMacAddress(ctypes.Structure):
  _fields_ = [
    ("bytes_", ctypes.c_uint8 * 6)
  ]


class HebiVector3f(ctypes.Structure):
  _fields_ = [
    ("x", ctypes.c_float),
    ("y", ctypes.c_float),
    ("z", ctypes.c_float),
  ]


class HebiQuaternionf(ctypes.Structure):
  _fields_ = [
    ("w", ctypes.c_float),
    ("x", ctypes.c_float),
    ("y", ctypes.c_float),
    ("z", ctypes.c_float),
  ]


class HebiRobotModelElementMetadata(ctypes.Structure):
  _anonymous_ = ("__element_struct_",)
  _fields_ = [
    ("struct_size_", ctypes.c_uint32),
    ("element_type_", ctypes.c_int),
    ("__element_struct_", _MetaDataUnionCTypes)
  ]


class HebiRobotModelElementTopology(ctypes.Structure):
  _fields_ = [
    ("element_index_", ctypes.c_int32),
    ("parent_index_", ctypes.c_int32),
    ("parent_output_", ctypes.c_int32),
    ("dof_location_", ctypes.c_int32),
    ("com_index_", ctypes.c_int32),
    ("output_index_", ctypes.c_int32),
    ("end_effector_index_", ctypes.c_int32)
  ]


class HebiHighResAngleStruct(ctypes.Structure):
  _fields_ = [
    ("revolutions_", ctypes.c_int64),
    ("offset_", ctypes.c_float)
  ]


class HebiIoBankPinStruct(ctypes.Structure):
  _anonymous_ = ("__io_union_",)
  _fields_ = [
    ("__io_union_", _IoPinUnion),
    ("stored_type_", ctypes.c_int)
  ]


class HebiCommandRef(ctypes.Structure):
  _fields_ = [
    ("message_bitfield_", ctypes.POINTER(ctypes.c_int32)),
    ("float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("high_res_angle_fields_", ctypes.POINTER(HebiHighResAngleStruct)),
    ("vector3f_fields_", ctypes.POINTER(HebiVector3f)),
    ("quaternionf_fields_", ctypes.POINTER(HebiQuaternionf)),
    ("uint64_fields_", ctypes.POINTER(ctypes.c_uint64)),
    ("enum_fields_", ctypes.POINTER(ctypes.c_int32)),
    ("bool_fields_", ctypes.POINTER(ctypes.c_bool)),
    ("numbered_float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("io_fields_", ctypes.POINTER(HebiIoBankPinStruct)),
    ("led_fields_", ctypes.POINTER(ctypes.c_uint32)),
    ("reserved_", ctypes.c_void_p)
  ]


class HebiCommandMetadata(ctypes.Structure):
  _fields_ = [
    ("float_field_count_", ctypes.c_uint32),
    ("high_res_angle_field_count_", ctypes.c_uint32),
    ("vector3f_field_count_", ctypes.c_uint32),
    ("quaternionf_field_count_", ctypes.c_uint32),
    ("uint64_field_count_", ctypes.c_uint32),
    ("enum_field_count_", ctypes.c_uint32),
    ("bool_field_count_", ctypes.c_uint32),
    ("numbered_float_field_count_", ctypes.c_uint32),
    ("io_field_count_", ctypes.c_uint32),
    ("led_field_count_", ctypes.c_uint32),
    ("string_field_count_", ctypes.c_uint32),
    ("flag_field_count_", ctypes.c_uint32),
    ("float_field_bitfield_offset_", ctypes.c_uint32),
    ("high_res_angle_field_bitfield_offset_", ctypes.c_uint32),
    ("vector3f_field_bitfield_offset_", ctypes.c_uint32),
    ("quaternionf_field_bitfield_offset_", ctypes.c_uint32),
    ("uint64_field_bitfield_offset_", ctypes.c_uint32),
    ("enum_field_bitfield_offset_", ctypes.c_uint32),
    ("bool_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_field_bitfield_offset_", ctypes.c_uint32),
    ("io_field_bitfield_offset_", ctypes.c_uint32),
    ("led_field_bitfield_offset_", ctypes.c_uint32),
    ("string_field_bitfield_offset_", ctypes.c_uint32),
    ("flag_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("numbered_float_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("message_bitfield_count_", ctypes.c_uint32),
  ]


class HebiFeedbackRef(ctypes.Structure):
  _fields_ = [
    ("message_bitfield_", ctypes.POINTER(ctypes.c_int32)),
    ("float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("high_res_angle_fields_", ctypes.POINTER(HebiHighResAngleStruct)),
    ("vector3f_fields_", ctypes.POINTER(HebiVector3f)),
    ("quaternionf_fields_", ctypes.POINTER(HebiQuaternionf)),
    ("uint64_fields_", ctypes.POINTER(ctypes.c_uint64)),
    ("enum_fields_", ctypes.POINTER(ctypes.c_int32)),
    ("bool_fields_", ctypes.POINTER(ctypes.c_bool)),
    ("numbered_float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("io_fields_", ctypes.POINTER(HebiIoBankPinStruct)),
    ("led_fields_", ctypes.POINTER(ctypes.c_uint32)),
    ("reserved_", ctypes.c_void_p)
  ]


class HebiFeedbackMetadata(ctypes.Structure):
  _fields_ = [
    ("float_field_count_", ctypes.c_uint32),
    ("high_res_angle_field_count_", ctypes.c_uint32),
    ("vector3f_field_count_", ctypes.c_uint32),
    ("quaternionf_field_count_", ctypes.c_uint32),
    ("uint64_field_count_", ctypes.c_uint32),
    ("enum_field_count_", ctypes.c_uint32),
    ("bool_field_count_", ctypes.c_uint32),
    ("numbered_float_field_count_", ctypes.c_uint32),
    ("io_field_count_", ctypes.c_uint32),
    ("led_field_count_", ctypes.c_uint32),
    ("string_field_count_", ctypes.c_uint32),
    ("flag_field_count_", ctypes.c_uint32),
    ("float_field_bitfield_offset_", ctypes.c_uint32),
    ("high_res_angle_field_bitfield_offset_", ctypes.c_uint32),
    ("vector3f_field_bitfield_offset_", ctypes.c_uint32),
    ("quaternionf_field_bitfield_offset_", ctypes.c_uint32),
    ("uint64_field_bitfield_offset_", ctypes.c_uint32),
    ("enum_field_bitfield_offset_", ctypes.c_uint32),
    ("bool_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_field_bitfield_offset_", ctypes.c_uint32),
    ("io_field_bitfield_offset_", ctypes.c_uint32),
    ("led_field_bitfield_offset_", ctypes.c_uint32),
    ("string_field_bitfield_offset_", ctypes.c_uint32),
    ("flag_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("numbered_float_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("message_bitfield_count_", ctypes.c_uint32),
  ]


class HebiInfoRef(ctypes.Structure):
  _fields_ = [
    ("message_bitfield_", ctypes.POINTER(ctypes.c_int32)),
    ("float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("high_res_angle_fields_", ctypes.POINTER(HebiHighResAngleStruct)),
    ("vector3f_fields_", ctypes.POINTER(HebiVector3f)),
    ("quaternionf_fields_", ctypes.POINTER(HebiQuaternionf)),
    ("uint64_fields_", ctypes.POINTER(ctypes.c_uint64)),
    ("enum_fields_", ctypes.POINTER(ctypes.c_int32)),
    ("bool_fields_", ctypes.POINTER(ctypes.c_bool)),
    ("numbered_float_fields_", ctypes.POINTER(ctypes.c_float)),
    ("io_fields_", ctypes.POINTER(HebiIoBankPinStruct)),
    ("led_fields_", ctypes.POINTER(ctypes.c_uint32)),
    ("reserved_", ctypes.c_void_p)
  ]


class HebiInfoMetadata(ctypes.Structure):
  _fields_ = [
    ("float_field_count_", ctypes.c_uint32),
    ("high_res_angle_field_count_", ctypes.c_uint32),
    ("vector3f_field_count_", ctypes.c_uint32),
    ("quaternionf_field_count_", ctypes.c_uint32),
    ("uint64_field_count_", ctypes.c_uint32),
    ("enum_field_count_", ctypes.c_uint32),
    ("bool_field_count_", ctypes.c_uint32),
    ("numbered_float_field_count_", ctypes.c_uint32),
    ("io_field_count_", ctypes.c_uint32),
    ("led_field_count_", ctypes.c_uint32),
    ("string_field_count_", ctypes.c_uint32),
    ("flag_field_count_", ctypes.c_uint32),
    ("float_field_bitfield_offset_", ctypes.c_uint32),
    ("high_res_angle_field_bitfield_offset_", ctypes.c_uint32),
    ("vector3f_field_bitfield_offset_", ctypes.c_uint32),
    ("quaternionf_field_bitfield_offset_", ctypes.c_uint32),
    ("uint64_field_bitfield_offset_", ctypes.c_uint32),
    ("enum_field_bitfield_offset_", ctypes.c_uint32),
    ("bool_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_field_bitfield_offset_", ctypes.c_uint32),
    ("io_field_bitfield_offset_", ctypes.c_uint32),
    ("led_field_bitfield_offset_", ctypes.c_uint32),
    ("string_field_bitfield_offset_", ctypes.c_uint32),
    ("flag_field_bitfield_offset_", ctypes.c_uint32),
    ("numbered_float_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("numbered_float_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_relative_offsets_", ctypes.POINTER(ctypes.c_uint32)),
    ("io_field_sizes_", ctypes.POINTER(ctypes.c_uint32)),
    ("message_bitfield_count_", ctypes.c_uint32),
  ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# ------------------------------------------------------------------------------

from ._internal.ffi import loader as __loader

from . import version
__version__ = str(version.py_version())


################################################################################
# Lookup API
################################################################################


from ._internal import lookup as _lookup_internal


class Lookup(object):
  """
  Maintains a registry of network-connected modules
  and returns Group objects to the user.
  """

  DEFAULT_TIMEOUT_MS = 500
  """
  The default timeout (in milliseconds)
  """

  __slots__ = ['__delegate']

  def __init__(self):
    self.__delegate = _lookup_internal.LookupDelegate.get_singleton()

  def __repr__(self):
    lookup_freq = self.lookup_frequency
    # We can't get lookup addresses from C API yet, so nothing to print re: that
    ret = 'lookup_frequency: {}\n\n'.format(lookup_freq)
    # Mobules table
    from ._internal.utils import lookup_table_string  
    return ret + lookup_table_string([entry for entry in self.entrylist])

  def __str__(self):
    lookup_freq = self.lookup_frequency
    modules = [entry for entry in self.entrylist]
    modules_string = None
    num_modules = len(modules)
    if num_modules > 1:
      modules_string = '{} modules'.format(num_modules)
    elif num_modules == 1:
      modules_string = '1 module'
    else:
      modules_string = 'no modules'
    return 'Lookup(lookup_frequency={}; {} found)'.format(lookup_freq, modules_string)

  @property
  def entrylist(self):
    """
    A list of discovered network connected modules.

    :return: The list of modules
    :rtype: EntryList
    """
    return self.__delegate.entrylist

  @property
  def lookup_frequency(self):
    return self.__delegate.lookup_frequency

  @lookup_frequency.setter
  def lookup_frequency(self, freq):
    self.__delegate.lookup_frequency = freq

  def get_group_from_names(self, families, names, timeout_ms=None):
    """
    Get a group from modules with the given names and families.

    If the families or names provided as input is only a single element,
    then that element is assumed to pair with each item in the other parameter.

    This is a blocking call which returns a Group with the given parameters.
    This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
    if a matching group cannot be constructed.

    :param families:
    :param names:
    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Lookup.DEFAULT_TIMEOUT_MS`.

    :return:
    """
    return self.__delegate.get_group_from_names(families, names, timeout_ms)

  def get_group_from_macs(self, addresses, timeout_ms=None):
    """
    Get a group from modules with the given mac addresses.

    This is a blocking call which returns a Group with the given parameters.
    This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
    if a matching group cannot be constructed.

    :param addresses:
    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Lookup.DEFAULT_TIMEOUT_MS`.

    :return:
    """
    return self.__delegate.get_group_from_macs(addresses, timeout_ms)

  def get_group_from_family(self, family, timeout_ms=None):
    """
    Get a group from all known modules with the given family.

    This is a blocking call which returns a Group with the given parameters.
    This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
    if a matching group cannot be constructed.

    :param family:
    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Lookup.DEFAULT_TIMEOUT_MS`.

    :return:
    """
    return self.__delegate.get_group_from_family(family, timeout_ms)

  def get_connected_group_from_name(self, family, name, timeout_ms=None):
    """
    Get a group from all modules known to connect to a module with
    the given name and family.

    This is a blocking call which returns a Group with the given parameters.
    This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
    if a matching group cannot be constructed.

    :param family:
    :param name:
    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Lookup.DEFAULT_TIMEOUT_MS`.

    :return:
    """
    return self.__delegate.get_connected_group_from_name(family, name, timeout_ms)

  def get_connected_group_from_mac(self, address, timeout_ms=None):
    """
    Get a group from all modules known to connect to a module with
    the given mac address.

    This is a blocking call which returns a Group with the given parameters.
    This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
    if a matching group cannot be constructed.

    :param address:
    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Lookup.DEFAULT_TIMEOUT_MS`.

    :return:
    """
    return self.__delegate.get_connected_group_from_mac(address, timeout_ms)


################################################################################
# Message Types
################################################################################


from ._internal.ffi._message_types import GroupCommand, GroupFeedback, GroupInfo
from ._internal.graphics import Color
from . import robot_model, trajectory, util

__all__ = ['Lookup', 'GroupCommand', 'GroupFeedback', 'GroupInfo', 'Color',
           'robot_model', 'trajectory', 'util']



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2017-2019 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------

from .ffi import enums as _enums


_feedback_scalars = [
  _enums.FeedbackFloatBoardTemperature,
  _enums.FeedbackFloatProcessorTemperature,
  _enums.FeedbackFloatVoltage,
  _enums.FeedbackFloatVelocity,
  _enums.FeedbackFloatEffort,
  _enums.FeedbackFloatVelocityCommand,
  _enums.FeedbackFloatEffortCommand,
  _enums.FeedbackFloatDeflection,
  _enums.FeedbackFloatDeflectionVelocity,
  _enums.FeedbackFloatMotorVelocity,
  _enums.FeedbackFloatMotorCurrent,
  _enums.FeedbackFloatMotorSensorTemperature,
  _enums.FeedbackFloatMotorWindingCurrent,
  _enums.FeedbackFloatMotorWindingTemperature,
  _enums.FeedbackFloatMotorHousingTemperature,
  _enums.FeedbackFloatBatteryLevel,
  _enums.FeedbackFloatPwmCommand,
  _enums.FeedbackHighResAnglePosition,
  _enums.FeedbackHighResAnglePositionCommand,
  _enums.FeedbackHighResAngleMotorPosition,
  _enums.FeedbackUInt64SequenceNumber,
  _enums.FeedbackUInt64ReceiveTime,
  _enums.FeedbackUInt64TransmitTime,
  _enums.FeedbackUInt64HardwareReceiveTime,
  _enums.FeedbackUInt64HardwareTransmitTime,
  _enums.FeedbackUInt64SenderId,
  _enums.FeedbackEnumTemperatureState,
  _enums.FeedbackEnumMstopState,
  _enums.FeedbackEnumPositionLimitState,
  _enums.FeedbackEnumVelocityLimitState,
  _enums.FeedbackEnumEffortLimitState,
  _enums.FeedbackEnumCommandLifetimeState,
  _enums.FeedbackEnumArQuality,
  _enums.FeedbackIoBankA,
  _enums.FeedbackIoBankB,
  _enums.FeedbackIoBankC,
  _enums.FeedbackIoBankD,
  _enums.FeedbackIoBankE,
  _enums.FeedbackIoBankF]


_feedback_scalars_map = {}


def __add_fbk_scalars_field(field):
  for alias in field.aliases:
    _feedback_scalars_map[alias] = field


def __populate_fbk_scalars_map():
  for entry in _feedback_scalars:

    field = entry.field_details

    if field is None:
      # TODO: THIS IS TEMPORARY. FIX THIS BY DEFINING FIELD_DETAILS FOR ALL FIELDS ABOVE
      continue

    if field.is_scalar:
      # Will be an instance of `MessageEnumTraits`
      __add_fbk_scalars_field(field)
    else:
      for sub_field in field.scalars.values():
        # Unspecified class type: will have all functionality of `MessageEnumTraits`
        __add_fbk_scalars_field(sub_field)


__populate_fbk_scalars_map()


def get_field_info(field_name):
  """
  Get the info object representing the given field name.

  The field binder is a lambda which accepts a group feedback instance and returns the input field name

  :param field_name:
  :return:
  """
  if field_name not in _feedback_scalars_map:
    raise KeyError(field_name)

  return _feedback_scalars_map[field_name]



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2017 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


import sys


def __default_out_stream(msg):
  print(msg)


def __default_err_stream(msg):
  sys.stderr.write(msg + '\n')


def __print_to_stream(msg, stream, prefix):
    if '\n' in msg:
      for entry in msg.splitlines():
        stream(prefix + entry)
    else:
      stream(prefix + msg)


__hebi_debug = False
__hebi_out_stream = __default_out_stream
__hebi_err_stream = __default_err_stream


def set_debug_mode(dbg):
  global __hebi_debug
  __hebi_debug = bool(dbg)


def debug_mode():
  return __hebi_debug


def debug_log(msg):
  if __hebi_debug:
    __print_to_stream(msg, __hebi_out_stream, '[HEBI Debug] ')


def warn_log(msg):
  __print_to_stream(msg, __hebi_err_stream, '[HEBI Warning] ')



from hebi._internal.utils import intern_string
import weakref

from .. import debug
from ..utils import Counter


################################################################################
# Function wrappers
################################################################################


class __DebugFunctionPrinterBypass(object):
  """
  Used to bypass printing C API calls while in Debug mode.
  This is to work around some known issues only encountered in debugging mode.
  """

  def __enter__(self):
    UnmanagedFunction.bypass_debug_printing = True
    return True

  def __exit__(self, exc_type, exc_val, exc_tb):
    UnmanagedFunction.bypass_debug_printing = False


bypass_debug_printing = __DebugFunctionPrinterBypass()


################################################################################
# Unmanaged class wrappers
################################################################################


class UnmanagedObject(object):
  """
  Base class for an object created by the HEBI C API.
  """

  __slots__ = ['_finalized', '_internal', '_lock', '_on_delete']

  def __init__(self, internal, on_delete=None):
    self._finalized = False
    self._internal = internal
    from threading import Lock
    self._lock = Lock()
    self._on_delete = on_delete

  @property
  def _as_parameter_(self):
    with self._lock:
      if self._finalized:
        import sys
        sys.stderr.write('Warning: Attempting to reference finalized HEBI object\n')
        raise RuntimeError('Object has already been finalized')
      return self._internal

  @property
  def finalized(self):
    with self._lock:
      return self._finalized

  def force_delete(self):
    with self._lock:
      if self._finalized:
        return
      if self._on_delete and self._internal:
        self._on_delete(self._internal)
      self._internal = None
      self._finalized = True

  def __del__(self):
    self.force_delete()


class UnmanagedSharedObject(object):
  """
  Base class for an object from the HEBI C API which requires
  reference counting.
  """

  __slots__ = ['_holder']

  def __init__(self, internal=None, on_delete=None, existing=None, isdummy=False):
    class Proxy(object):
      def __init__(self, internal=None, on_delete=None, proxy=None):
        if proxy is not None:
          with proxy._lock:
            if proxy._refcount.count == 0:
              raise RuntimeError('Object has already been finalized')
            self._lock = proxy._lock
            self._internal = proxy._internal
            self._on_delete = proxy._on_delete
            proxy._refcount.increment()
            self._refcount = proxy._refcount
        else:
          from threading import Lock
          self._lock = Lock()
          self._internal = internal
          self._refcount = Counter()
          self._on_delete = on_delete
          
      def __del__(self):
        with self._lock:
          if self._refcount.count > 0:
            # Object has not been explicitly deleted
            self._refcount.decrement()
          if self._refcount.count == 0 and self._on_delete is not None:
            self._on_delete(self._internal)

      def force_delete(self):
        with self._lock:
          if self._refcount.count == 0:
            # Already deleted - return
            return
          self._refcount.clear()
          if self._on_delete is not None:
            self._on_delete(self._internal)
            self._on_delete = None
          self._internal = None

      def get(self):
        with self._lock:
          if self._refcount.count == 0:
            raise RuntimeError('Object has already been finalized')
          return self._internal

    if existing is not None:
      if not isinstance(existing, UnmanagedSharedObject):
        raise TypeError('existing parameter must be an UnmanagedSharedObject instance')
      self._holder = Proxy(proxy=existing._holder)
    else:
      self._holder = Proxy(internal=internal, on_delete=on_delete)

  @property
  def _as_parameter_(self):
    return self._holder.get()


class WeakReferenceContainer(object):
  """
  Small wrapper around a weak reference. For internal use - do not use directly.
  """

  __slots__ = ['_weak_ref']

  def _get_ref(self):
    ref = self._weak_ref()
    if ref is not None:
      return ref
    raise RuntimeError('Reference no longer valid due to finalization')

  def __init__(self, ref):
    self._weak_ref = weakref.ref(ref)


################################################################################
# Enum wrappers
################################################################################


class FieldDetails(object):
  """
  TODO: Document
  """

  __slots__ = ['_units', '_camel_case', '_pascal_case', '_snake_case', '_is_scalar']

  def __init__(self, units, camel, pascal, snake, is_scalar):
    if units is None:
      self._units = None
    else:
      self._units = intern_string(units)

    self._camel_case = intern_string(camel)
    self._pascal_case = intern_string(pascal)
    self._snake_case = intern_string(snake)
    self._is_scalar = is_scalar

  @property
  def aliases(self):
    return {self._snake_case, self._pascal_case, self._camel_case}

  @property
  def units(self):
    return self._units

  @property
  def is_scalar(self):
    return self._is_scalar

  @property
  def camel_case(self):
    return self._camel_case

  @property
  def pascal_case(self):
    return self._pascal_case

  @property
  def snake_case(self):
    return self._snake_case

  def get_field(self, message_obj):
    """
    Retrieve the field value(s) from the given message object.

    The message object can be a single or group message (e.g.: Feedback, GroupFeedback, Command, GroupCommand, etc)

    :param message_obj: Message object which has the field represented by `self.snake_case`

    :raises AttributeError: If the field does not exist on the provided object
    """
    return getattr(message_obj, self._snake_case)


class NumberedFieldDetails(object):
  __slots__ = ['_getter_factory', '_units', '_base_camel_case', '_base_pascal_case', '_base_snake_case', '_numbered_fields']

  def __init__(self, units, camel, pascal, snake, getter_factory, numbered_fields):
    if units is None:
      self._units = None
    else:
      self._units = intern_string(units)

    self._base_camel_case = snake
    self._base_pascal_case = pascal
    self._base_snake_case = camel
    self._getter_factory = getter_factory
    self._numbered_fields = numbered_fields

  @property
  def units(self):
    return self._units

  @property
  def is_scalar(self):
    return False

  @property
  def scalars(self):
    """
    Retrieve all scalar fields (in a dictionary) encapsulated in this field.
    All elements within the dictionary will be a class similar to :class:`FieldDetails`.
    """
    num_fields = len(self._numbered_fields)
    ret = dict()

    base_camel_case = self._base_camel_case
    base_pascal_case = self._base_pascal_case
    base_snake_case = self._base_snake_case
    base_units = self._units

    for i in range(num_fields):
      sub_field = self._numbered_fields[i]
      sub_field_getter = self._getter_factory(sub_field)
      sub_field_str = str(sub_field)

      class Subfield(object):
        __slots__ = []

        def __init__(self):
          pass

        @property
        def aliases(self):
          return {base_camel_case + sub_field_str, base_pascal_case + sub_field_str, base_snake_case + sub_field_str}

        @property
        def units(self):
          return base_units

        def get_field(self, message_obj):
          return sub_field_getter(message_obj)


      ret[sub_field] = Subfield()

    return ret


class EnumTraits(object):
  """
  TODO: Document
  """

  __slots__ = ['_as_parameter_', '_name', '_value', '_type']

  def __init__(self, value, name):
    self._value = value
    self._name = intern_string(name)
    self._as_parameter_ = value
    self._type = None

  @property
  def value(self):
    return self._value

  @property
  def name(self):
    return self._name

  @property
  def type(self):
    return self._type

  def __int__(self):
    return self._value

  def __hash__(self):
    return self._value + (31*hash(self._name))

  def __eq__(self, value):
    if type(value) is int:
      return value == self._value
    elif isinstance(value, EnumTraits):
      return (value._name is self._name) and (value._value == self._value)
    elif type(value) is str:
      return intern_string(value) is self._name
    else:
      return value == self._value

  def __ne__(self, value):
    if type(value) is int:
      return value != self._value
    elif isinstance(value, EnumTraits):
      return (value._name is not self._name) or (value._value != self._value)
    elif type(value) is str:
      return value is not self._name
    else:
      return value != self._value

  def __str__(self):
    return self._name

  def __repr__(self):
    return '{} (value={}, type={})'.format(self._name, self._value, self.type)


class MessageEnumTraits(EnumTraits):
  """
  TODO: Document
  """

  __slots__ = ['__allow_broadcast', '__products', '__substrates', '__not_broadcastable_reason', '__field_details']

  def __init__(self, value, name, field_details=None, allow_broadcast=True, not_bcastable_reason='', yields_dict=None):
    super(MessageEnumTraits, self).__init__(value, name)
    self.__allow_broadcast = allow_broadcast
    self.__not_broadcastable_reason = not_bcastable_reason
    self.__products = yields_dict
    self.__field_details = field_details

    if yields_dict is not None:
      substrate = dict()
      for key in yields_dict.keys():
        value = yields_dict[key].lower()
        substrate[value] = key
      self.__substrates = substrate
    else:
      self.__substrates = None

  @property
  def allow_broadcast(self):
    """
    Determines whether the given enum has the ability to broadcast the
    same value to all modules in a group
    """
    return self.__allow_broadcast

  @property
  def substrates(self):
    return self.__substrates

  @property
  def products(self):
    return self.__products

  @property
  def not_broadcastable_reason(self):
    if self.allow_broadcast:
      return ''
    return self.__not_broadcastable_reason

  @property
  def field_details(self):
    return self.__field_details

  def __repr__(self):
    if self.__allow_broadcast:
      return '{} (value={}, type={}), broadcastable'.format(self._name, self._value, self.type)
    return '{} (value={}, type={}), not broadcastable ({})'.format(self._name, self._value, self.type, self.__not_broadcastable_reason)


class EnumType(object):
  __slots__ = ["__name", "__values", "__int_to_value"]

  def __init__(self, name : str, values : list):
    self.__name = name

    int_to_value = dict()

    for value in values:
      if not isinstance(value, EnumTraits):
        raise TypeError("values must all be EnumTraits")
      else:
        value._type = self
        int_to_value[int(value)] = value

    self.__int_to_value = int_to_value
    self.__values = values

  @property
  def name(self):
    return self.__name

  @property
  def values(self):
    return self.__values

  def get_enum_value_by_int(self, value : int):
    if value in self.__int_to_value:
      return self.__int_to_value[value]
    else:
      raise KeyError("'{}' is not a valid enum value".format(value))



import ctypes
from .ctypes_defs import HebiMacAddress, HebiVector3f, HebiQuaternionf, HebiRobotModelElementMetadata, HebiRobotModelElementTopology, HebiHighResAngleStruct, HebiIoBankPinStruct, HebiCommandRef, HebiCommandMetadata, HebiFeedbackRef, HebiFeedbackMetadata, HebiInfoRef, HebiInfoMetadata


def populate_hebi_funcs(lib):
  """
  Populates all API functions from the HEBI Core library
  """
  
  if hasattr(lib, 'hebiLookupCreate'):
    lib_func = getattr(lib, 'hebiLookupCreate')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_char_p),ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiLookupCreate')

  if hasattr(lib, 'hebiLookupRelease'):
    lib_func = getattr(lib, 'hebiLookupRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiLookupRelease')

  if hasattr(lib, 'hebiLookupSetLookupFrequencyHz'):
    lib_func = getattr(lib, 'hebiLookupSetLookupFrequencyHz')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLookupSetLookupFrequencyHz')

  if hasattr(lib, 'hebiLookupGetLookupFrequencyHz'):
    lib_func = getattr(lib, 'hebiLookupGetLookupFrequencyHz')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_double
  else:
    raise RuntimeWarning('Could not load function hebiLookupGetLookupFrequencyHz')

  if hasattr(lib, 'hebiCreateLookupEntryList'):
    lib_func = getattr(lib, 'hebiCreateLookupEntryList')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiCreateLookupEntryList')

  if hasattr(lib, 'hebiLookupEntryListGetSize'):
    lib_func = getattr(lib, 'hebiLookupEntryListGetSize')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiLookupEntryListGetSize')

  if hasattr(lib, 'hebiLookupEntryListGetName'):
    lib_func = getattr(lib, 'hebiLookupEntryListGetName')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLookupEntryListGetName')

  if hasattr(lib, 'hebiLookupEntryListGetFamily'):
    lib_func = getattr(lib, 'hebiLookupEntryListGetFamily')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLookupEntryListGetFamily')

  if hasattr(lib, 'hebiLookupEntryListGetMacAddress'):
    lib_func = getattr(lib, 'hebiLookupEntryListGetMacAddress')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t,ctypes.POINTER(HebiMacAddress)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLookupEntryListGetMacAddress')

  if hasattr(lib, 'hebiLookupEntryListRelease'):
    lib_func = getattr(lib, 'hebiLookupEntryListRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiLookupEntryListRelease')

  if hasattr(lib, 'hebiGroupCreateImitation'):
    lib_func = getattr(lib, 'hebiGroupCreateImitation')
    lib_func.argtypes = [ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateImitation')

  if hasattr(lib, 'hebiGroupCreateFromMacs'):
    lib_func = getattr(lib, 'hebiGroupCreateFromMacs')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(ctypes.POINTER(HebiMacAddress)),ctypes.c_size_t,ctypes.c_int32]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateFromMacs')

  if hasattr(lib, 'hebiGroupCreateFromNames'):
    lib_func = getattr(lib, 'hebiGroupCreateFromNames')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(ctypes.c_char_p),ctypes.c_size_t,ctypes.POINTER(ctypes.c_char_p),ctypes.c_size_t,ctypes.c_int32]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateFromNames')

  if hasattr(lib, 'hebiGroupCreateFromFamily'):
    lib_func = getattr(lib, 'hebiGroupCreateFromFamily')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateFromFamily')

  if hasattr(lib, 'hebiGroupCreateConnectedFromMac'):
    lib_func = getattr(lib, 'hebiGroupCreateConnectedFromMac')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(HebiMacAddress),ctypes.c_int32]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateConnectedFromMac')

  if hasattr(lib, 'hebiGroupCreateConnectedFromName'):
    lib_func = getattr(lib, 'hebiGroupCreateConnectedFromName')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p,ctypes.c_char_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCreateConnectedFromName')

  if hasattr(lib, 'hebiGroupGetSize'):
    lib_func = getattr(lib, 'hebiGroupGetSize')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiGroupGetSize')

  if hasattr(lib, 'hebiGroupSendCommandWithAcknowledgement'):
    lib_func = getattr(lib, 'hebiGroupSendCommandWithAcknowledgement')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupSendCommandWithAcknowledgement')

  if hasattr(lib, 'hebiGroupSendCommand'):
    lib_func = getattr(lib, 'hebiGroupSendCommand')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupSendCommand')

  if hasattr(lib, 'hebiGroupSetCommandLifetime'):
    lib_func = getattr(lib, 'hebiGroupSetCommandLifetime')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupSetCommandLifetime')

  if hasattr(lib, 'hebiGroupGetCommandLifetime'):
    lib_func = getattr(lib, 'hebiGroupGetCommandLifetime')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_int32
  else:
    raise RuntimeWarning('Could not load function hebiGroupGetCommandLifetime')

  if hasattr(lib, 'hebiGroupSetFeedbackFrequencyHz'):
    lib_func = getattr(lib, 'hebiGroupSetFeedbackFrequencyHz')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_float]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupSetFeedbackFrequencyHz')

  if hasattr(lib, 'hebiGroupGetFeedbackFrequencyHz'):
    lib_func = getattr(lib, 'hebiGroupGetFeedbackFrequencyHz')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_float
  else:
    raise RuntimeWarning('Could not load function hebiGroupGetFeedbackFrequencyHz')

  if hasattr(lib, 'hebiGroupRegisterFeedbackHandler'):
    lib_func = getattr(lib, 'hebiGroupRegisterFeedbackHandler')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.CFUNCTYPE(None,ctypes.c_void_p,ctypes.c_void_p),ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupRegisterFeedbackHandler')

  if hasattr(lib, 'hebiGroupClearFeedbackHandlers'):
    lib_func = getattr(lib, 'hebiGroupClearFeedbackHandlers')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupClearFeedbackHandlers')

  if hasattr(lib, 'hebiGroupSendFeedbackRequest'):
    lib_func = getattr(lib, 'hebiGroupSendFeedbackRequest')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupSendFeedbackRequest')

  if hasattr(lib, 'hebiGroupGetNextFeedback'):
    lib_func = getattr(lib, 'hebiGroupGetNextFeedback')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupGetNextFeedback')

  if hasattr(lib, 'hebiGroupRequestInfo'):
    lib_func = getattr(lib, 'hebiGroupRequestInfo')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_int32]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupRequestInfo')

  if hasattr(lib, 'hebiGroupStartLog'):
    lib_func = getattr(lib, 'hebiGroupStartLog')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p,ctypes.c_char_p,ctypes.POINTER(ctypes.c_void_p)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupStartLog')

  if hasattr(lib, 'hebiGroupStopLog'):
    lib_func = getattr(lib, 'hebiGroupStopLog')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupStopLog')

  if hasattr(lib, 'hebiGroupRelease'):
    lib_func = getattr(lib, 'hebiGroupRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupRelease')

  if hasattr(lib, 'hebiGroupCommandCreate'):
    lib_func = getattr(lib, 'hebiGroupCommandCreate')
    lib_func.argtypes = [ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandCreate')

  if hasattr(lib, 'hebiGroupCommandGetSize'):
    lib_func = getattr(lib, 'hebiGroupCommandGetSize')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandGetSize')

  if hasattr(lib, 'hebiGroupCommandReadGains'):
    lib_func = getattr(lib, 'hebiGroupCommandReadGains')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandReadGains')

  if hasattr(lib, 'hebiGroupCommandWriteGains'):
    lib_func = getattr(lib, 'hebiGroupCommandWriteGains')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandWriteGains')

  if hasattr(lib, 'hebiGroupCommandReadSafetyParameters'):
    lib_func = getattr(lib, 'hebiGroupCommandReadSafetyParameters')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandReadSafetyParameters')

  if hasattr(lib, 'hebiGroupCommandWriteSafetyParameters'):
    lib_func = getattr(lib, 'hebiGroupCommandWriteSafetyParameters')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandWriteSafetyParameters')

  if hasattr(lib, 'hebiGroupCommandGetModuleCommand'):
    lib_func = getattr(lib, 'hebiGroupCommandGetModuleCommand')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandGetModuleCommand')

  if hasattr(lib, 'hebiGroupCommandCopy'):
    lib_func = getattr(lib, 'hebiGroupCommandCopy')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandCopy')

  if hasattr(lib, 'hebiGroupCommandClear'):
    lib_func = getattr(lib, 'hebiGroupCommandClear')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandClear')

  if hasattr(lib, 'hebiGroupCommandRelease'):
    lib_func = getattr(lib, 'hebiGroupCommandRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupCommandRelease')

  if hasattr(lib, 'hebiGroupFeedbackCreate'):
    lib_func = getattr(lib, 'hebiGroupFeedbackCreate')
    lib_func.argtypes = [ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackCreate')

  if hasattr(lib, 'hebiGroupFeedbackGetSize'):
    lib_func = getattr(lib, 'hebiGroupFeedbackGetSize')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackGetSize')

  if hasattr(lib, 'hebiGroupFeedbackGetModuleFeedback'):
    lib_func = getattr(lib, 'hebiGroupFeedbackGetModuleFeedback')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackGetModuleFeedback')

  if hasattr(lib, 'hebiGroupFeedbackCopy'):
    lib_func = getattr(lib, 'hebiGroupFeedbackCopy')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackCopy')

  if hasattr(lib, 'hebiGroupFeedbackClear'):
    lib_func = getattr(lib, 'hebiGroupFeedbackClear')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackClear')

  if hasattr(lib, 'hebiGroupFeedbackRelease'):
    lib_func = getattr(lib, 'hebiGroupFeedbackRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupFeedbackRelease')

  if hasattr(lib, 'hebiGroupInfoCreate'):
    lib_func = getattr(lib, 'hebiGroupInfoCreate')
    lib_func.argtypes = [ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoCreate')

  if hasattr(lib, 'hebiGroupInfoGetSize'):
    lib_func = getattr(lib, 'hebiGroupInfoGetSize')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoGetSize')

  if hasattr(lib, 'hebiGroupInfoWriteGains'):
    lib_func = getattr(lib, 'hebiGroupInfoWriteGains')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoWriteGains')

  if hasattr(lib, 'hebiGroupInfoWriteSafetyParameters'):
    lib_func = getattr(lib, 'hebiGroupInfoWriteSafetyParameters')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoWriteSafetyParameters')

  if hasattr(lib, 'hebiGroupInfoGetModuleInfo'):
    lib_func = getattr(lib, 'hebiGroupInfoGetModuleInfo')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoGetModuleInfo')

  if hasattr(lib, 'hebiGroupInfoCopy'):
    lib_func = getattr(lib, 'hebiGroupInfoCopy')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoCopy')

  if hasattr(lib, 'hebiGroupInfoClear'):
    lib_func = getattr(lib, 'hebiGroupInfoClear')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoClear')

  if hasattr(lib, 'hebiGroupInfoRelease'):
    lib_func = getattr(lib, 'hebiGroupInfoRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiGroupInfoRelease')

  if hasattr(lib, 'hebiCommandGetString'):
    lib_func = getattr(lib, 'hebiCommandGetString')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiCommandGetString')

  if hasattr(lib, 'hebiCommandSetString'):
    lib_func = getattr(lib, 'hebiCommandSetString')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiCommandSetString')

  if hasattr(lib, 'hebiCommandGetReference'):
    lib_func = getattr(lib, 'hebiCommandGetReference')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(HebiCommandRef)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiCommandGetReference')

  if hasattr(lib, 'hebiCommandGetMetadata'):
    lib_func = getattr(lib, 'hebiCommandGetMetadata')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandMetadata)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiCommandGetMetadata')

  if hasattr(lib, 'hebiFeedbackGetReference'):
    lib_func = getattr(lib, 'hebiFeedbackGetReference')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(HebiFeedbackRef)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiFeedbackGetReference')

  if hasattr(lib, 'hebiFeedbackGetMetadata'):
    lib_func = getattr(lib, 'hebiFeedbackGetMetadata')
    lib_func.argtypes = [ctypes.POINTER(HebiFeedbackMetadata)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiFeedbackGetMetadata')

  if hasattr(lib, 'hebiInfoGetString'):
    lib_func = getattr(lib, 'hebiInfoGetString')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiInfoGetString')

  if hasattr(lib, 'hebiInfoGetReference'):
    lib_func = getattr(lib, 'hebiInfoGetReference')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(HebiInfoRef)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiInfoGetReference')

  if hasattr(lib, 'hebiInfoGetMetadata'):
    lib_func = getattr(lib, 'hebiInfoGetMetadata')
    lib_func.argtypes = [ctypes.POINTER(HebiInfoMetadata)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiInfoGetMetadata')

  if hasattr(lib, 'hebiRobotModelElementCreateJoint'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateJoint')
    lib_func.argtypes = [ctypes.c_int]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateJoint')

  if hasattr(lib, 'hebiRobotModelElementCreateRigidBody'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateRigidBody')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.c_double,ctypes.c_size_t,ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateRigidBody')

  if hasattr(lib, 'hebiRobotModelElementCreateEndEffector'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateEndEffector')
    lib_func.argtypes = [ctypes.c_int,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.c_double,ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateEndEffector')

  if hasattr(lib, 'hebiRobotModelElementCreateActuator'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateActuator')
    lib_func.argtypes = [ctypes.c_int]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateActuator')

  if hasattr(lib, 'hebiRobotModelElementCreateBracket'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateBracket')
    lib_func.argtypes = [ctypes.c_int]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateBracket')

  if hasattr(lib, 'hebiRobotModelElementCreateLink'):
    lib_func = getattr(lib, 'hebiRobotModelElementCreateLink')
    lib_func.argtypes = [ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_double,ctypes.c_double]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementCreateLink')

  if hasattr(lib, 'hebiRobotModelElementRelease'):
    lib_func = getattr(lib, 'hebiRobotModelElementRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelElementRelease')

  if hasattr(lib, 'hebiRobotModelImport'):
    lib_func = getattr(lib, 'hebiRobotModelImport')
    lib_func.argtypes = [ctypes.c_char_p]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelImport')

  if hasattr(lib, 'hebiRobotModelImportBuffer'):
    lib_func = getattr(lib, 'hebiRobotModelImportBuffer')
    lib_func.argtypes = [ctypes.c_char_p,ctypes.c_size_t]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelImportBuffer')

  if hasattr(lib, 'hebiRobotModelGetImportError'):
    lib_func = getattr(lib, 'hebiRobotModelGetImportError')
    lib_func.argtypes = []
    lib_func.restype = ctypes.c_char_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetImportError')

  if hasattr(lib, 'hebiRobotModelGetImportWarningCount'):
    lib_func = getattr(lib, 'hebiRobotModelGetImportWarningCount')
    lib_func.argtypes = []
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetImportWarningCount')

  if hasattr(lib, 'hebiRobotModelGetImportWarning'):
    lib_func = getattr(lib, 'hebiRobotModelGetImportWarning')
    lib_func.argtypes = [ctypes.c_size_t]
    lib_func.restype = ctypes.c_char_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetImportWarning')

  if hasattr(lib, 'hebiRobotModelCreate'):
    lib_func = getattr(lib, 'hebiRobotModelCreate')
    lib_func.argtypes = []
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelCreate')

  if hasattr(lib, 'hebiRobotModelSetBaseFrame'):
    lib_func = getattr(lib, 'hebiRobotModelSetBaseFrame')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelSetBaseFrame')

  if hasattr(lib, 'hebiRobotModelGetBaseFrame'):
    lib_func = getattr(lib, 'hebiRobotModelGetBaseFrame')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetBaseFrame')

  if hasattr(lib, 'hebiRobotModelGetNumberOfFrames'):
    lib_func = getattr(lib, 'hebiRobotModelGetNumberOfFrames')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetNumberOfFrames')

  if hasattr(lib, 'hebiRobotModelGetNumberOfDoFs'):
    lib_func = getattr(lib, 'hebiRobotModelGetNumberOfDoFs')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetNumberOfDoFs')

  if hasattr(lib, 'hebiRobotModelGetNumberOfElements'):
    lib_func = getattr(lib, 'hebiRobotModelGetNumberOfElements')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetNumberOfElements')

  if hasattr(lib, 'hebiRobotModelGetElementMetadata'):
    lib_func = getattr(lib, 'hebiRobotModelGetElementMetadata')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_size_t,ctypes.POINTER(HebiRobotModelElementMetadata)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetElementMetadata')

  if hasattr(lib, 'hebiRobotModelAdd'):
    lib_func = getattr(lib, 'hebiRobotModelAdd')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_size_t,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelAdd')

  if hasattr(lib, 'hebiRobotModelGetForwardKinematics'):
    lib_func = getattr(lib, 'hebiRobotModelGetForwardKinematics')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetForwardKinematics')

  if hasattr(lib, 'hebiRobotModelGetJacobians'):
    lib_func = getattr(lib, 'hebiRobotModelGetJacobians')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetJacobians')

  if hasattr(lib, 'hebiRobotModelGetMasses'):
    lib_func = getattr(lib, 'hebiRobotModelGetMasses')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.POINTER(ctypes.c_double)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetMasses')

  if hasattr(lib, 'hebiRobotModelGetTreeTopology'):
    lib_func = getattr(lib, 'hebiRobotModelGetTreeTopology')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.POINTER(HebiRobotModelElementTopology)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelGetTreeTopology')

  if hasattr(lib, 'hebiRobotModelRelease'):
    lib_func = getattr(lib, 'hebiRobotModelRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiRobotModelRelease')

  if hasattr(lib, 'hebiIKCreate'):
    lib_func = getattr(lib, 'hebiIKCreate')
    lib_func.argtypes = []
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiIKCreate')

  if hasattr(lib, 'hebiIKAddObjectiveEndEffectorPosition'):
    lib_func = getattr(lib, 'hebiIKAddObjectiveEndEffectorPosition')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.c_size_t,ctypes.c_double,ctypes.c_double,ctypes.c_double]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKAddObjectiveEndEffectorPosition')

  if hasattr(lib, 'hebiIKAddObjectiveEndEffectorSO3'):
    lib_func = getattr(lib, 'hebiIKAddObjectiveEndEffectorSO3')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.c_size_t,ctypes.POINTER(ctypes.c_double),ctypes.c_int]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKAddObjectiveEndEffectorSO3')

  if hasattr(lib, 'hebiIKAddObjectiveEndEffectorTipAxis'):
    lib_func = getattr(lib, 'hebiIKAddObjectiveEndEffectorTipAxis')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.c_size_t,ctypes.c_double,ctypes.c_double,ctypes.c_double]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKAddObjectiveEndEffectorTipAxis')

  if hasattr(lib, 'hebiIKAddConstraintJointAngles'):
    lib_func = getattr(lib, 'hebiIKAddConstraintJointAngles')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.c_size_t,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKAddConstraintJointAngles')

  if hasattr(lib, 'hebiIKAddObjectiveCustom'):
    lib_func = getattr(lib, 'hebiIKAddObjectiveCustom')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.c_size_t,ctypes.CFUNCTYPE(None,ctypes.c_void_p,ctypes.c_size_t,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double)),ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKAddObjectiveCustom')

  if hasattr(lib, 'hebiIKClearAll'):
    lib_func = getattr(lib, 'hebiIKClearAll')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiIKClearAll')

  if hasattr(lib, 'hebiIKSolve'):
    lib_func = getattr(lib, 'hebiIKSolve')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiIKSolve')

  if hasattr(lib, 'hebiIKRelease'):
    lib_func = getattr(lib, 'hebiIKRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiIKRelease')

  if hasattr(lib, 'hebiTrajectoryCreateUnconstrainedQp'):
    lib_func = getattr(lib, 'hebiTrajectoryCreateUnconstrainedQp')
    lib_func.argtypes = [ctypes.c_size_t,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double)]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiTrajectoryCreateUnconstrainedQp')

  if hasattr(lib, 'hebiTrajectoryRelease'):
    lib_func = getattr(lib, 'hebiTrajectoryRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiTrajectoryRelease')

  if hasattr(lib, 'hebiTrajectoryGetDuration'):
    lib_func = getattr(lib, 'hebiTrajectoryGetDuration')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_double
  else:
    raise RuntimeWarning('Could not load function hebiTrajectoryGetDuration')

  if hasattr(lib, 'hebiTrajectoryGetState'):
    lib_func = getattr(lib, 'hebiTrajectoryGetState')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_double,ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double),ctypes.POINTER(ctypes.c_double)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiTrajectoryGetState')

  if hasattr(lib, 'hebiLogFileRelease'):
    lib_func = getattr(lib, 'hebiLogFileRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiLogFileRelease')

  if hasattr(lib, 'hebiLogFileGetFileName'):
    lib_func = getattr(lib, 'hebiLogFileGetFileName')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLogFileGetFileName')

  if hasattr(lib, 'hebiLogFileOpen'):
    lib_func = getattr(lib, 'hebiLogFileOpen')
    lib_func.argtypes = [ctypes.c_char_p]
    lib_func.restype = ctypes.c_void_p
  else:
    raise RuntimeWarning('Could not load function hebiLogFileOpen')

  if hasattr(lib, 'hebiLogFileGetNumberOfModules'):
    lib_func = getattr(lib, 'hebiLogFileGetNumberOfModules')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = ctypes.c_size_t
  else:
    raise RuntimeWarning('Could not load function hebiLogFileGetNumberOfModules')

  if hasattr(lib, 'hebiLogFileGetNextFeedback'):
    lib_func = getattr(lib, 'hebiLogFileGetNextFeedback')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiLogFileGetNextFeedback')

  if hasattr(lib, 'hebiStringGetString'):
    lib_func = getattr(lib, 'hebiStringGetString')
    lib_func.argtypes = [ctypes.c_void_p,ctypes.c_char_p,ctypes.POINTER(ctypes.c_size_t)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiStringGetString')

  if hasattr(lib, 'hebiStringRelease'):
    lib_func = getattr(lib, 'hebiStringRelease')
    lib_func.argtypes = [ctypes.c_void_p]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiStringRelease')

  if hasattr(lib, 'hebiSafetyParametersGetLastError'):
    lib_func = getattr(lib, 'hebiSafetyParametersGetLastError')
    lib_func.argtypes = []
    lib_func.restype = ctypes.c_char_p
  else:
    raise RuntimeWarning('Could not load function hebiSafetyParametersGetLastError')

  if hasattr(lib, 'hebiGetLibraryVersion'):
    lib_func = getattr(lib, 'hebiGetLibraryVersion')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(ctypes.c_int32)]
    lib_func.restype = ctypes.c_int
  else:
    raise RuntimeWarning('Could not load function hebiGetLibraryVersion')

  if hasattr(lib, 'hebiCleanup'):
    lib_func = getattr(lib, 'hebiCleanup')
    lib_func.argtypes = []
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hebiCleanup')


def populate_hebi_wrapper_funcs(lib):
  """
  Populates all API functions from the HEBI Core wrapper library
  """
  
  if hasattr(lib, 'hwInitialize'):
    lib_func = getattr(lib, 'hwInitialize')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandMetadata),ctypes.POINTER(HebiFeedbackMetadata),ctypes.POINTER(HebiInfoMetadata)]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInitialize')

  if hasattr(lib, 'hwCommandSetFloat'):
    lib_func = getattr(lib, 'hwCommandSetFloat')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_float),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetFloat')

  if hasattr(lib, 'hwCommandSetHighResAngle'):
    lib_func = getattr(lib, 'hwCommandSetHighResAngle')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_double),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetHighResAngle')

  if hasattr(lib, 'hwCommandSetEnum'):
    lib_func = getattr(lib, 'hwCommandSetEnum')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_int32),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetEnum')

  if hasattr(lib, 'hwCommandSetBool'):
    lib_func = getattr(lib, 'hwCommandSetBool')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_int32),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetBool')

  if hasattr(lib, 'hwCommandSetNumberedFloat'):
    lib_func = getattr(lib, 'hwCommandSetNumberedFloat')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_float),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetNumberedFloat')

  if hasattr(lib, 'hwCommandSetIoPin'):
    lib_func = getattr(lib, 'hwCommandSetIoPin')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(HebiIoBankPinStruct),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetIoPin')

  if hasattr(lib, 'hwCommandSetIoPinInt'):
    lib_func = getattr(lib, 'hwCommandSetIoPinInt')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_int64),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetIoPinInt')

  if hasattr(lib, 'hwCommandSetIoPinFloat'):
    lib_func = getattr(lib, 'hwCommandSetIoPinFloat')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_float),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetIoPinFloat')

  if hasattr(lib, 'hwCommandSetLed'):
    lib_func = getattr(lib, 'hwCommandSetLed')
    lib_func.argtypes = [ctypes.POINTER(HebiCommandRef),ctypes.POINTER(ctypes.c_int32),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandSetLed')

  if hasattr(lib, 'hwCommandGetFloat'):
    lib_func = getattr(lib, 'hwCommandGetFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetFloat')

  if hasattr(lib, 'hwCommandGetHighResAngle'):
    lib_func = getattr(lib, 'hwCommandGetHighResAngle')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_double),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetHighResAngle')

  if hasattr(lib, 'hwCommandGetVector3f'):
    lib_func = getattr(lib, 'hwCommandGetVector3f')
    lib_func.argtypes = [ctypes.POINTER(HebiVector3f),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetVector3f')

  if hasattr(lib, 'hwCommandGetQuaternionf'):
    lib_func = getattr(lib, 'hwCommandGetQuaternionf')
    lib_func.argtypes = [ctypes.POINTER(HebiQuaternionf),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetQuaternionf')

  if hasattr(lib, 'hwCommandGetUInt64'):
    lib_func = getattr(lib, 'hwCommandGetUInt64')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_uint64),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetUInt64')

  if hasattr(lib, 'hwCommandGetEnum'):
    lib_func = getattr(lib, 'hwCommandGetEnum')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetEnum')

  if hasattr(lib, 'hwCommandGetBool'):
    lib_func = getattr(lib, 'hwCommandGetBool')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetBool')

  if hasattr(lib, 'hwCommandGetNumberedFloat'):
    lib_func = getattr(lib, 'hwCommandGetNumberedFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetNumberedFloat')

  if hasattr(lib, 'hwCommandGetIoPin'):
    lib_func = getattr(lib, 'hwCommandGetIoPin')
    lib_func.argtypes = [ctypes.POINTER(HebiIoBankPinStruct),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetIoPin')

  if hasattr(lib, 'hwCommandGetIoPinInt'):
    lib_func = getattr(lib, 'hwCommandGetIoPinInt')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int64),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetIoPinInt')

  if hasattr(lib, 'hwCommandGetIoPinFloat'):
    lib_func = getattr(lib, 'hwCommandGetIoPinFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetIoPinFloat')

  if hasattr(lib, 'hwCommandGetLed'):
    lib_func = getattr(lib, 'hwCommandGetLed')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandGetLed')

  if hasattr(lib, 'hwCommandHasField'):
    lib_func = getattr(lib, 'hwCommandHasField')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandHasField')

  if hasattr(lib, 'hwCommandHasIoPinInt'):
    lib_func = getattr(lib, 'hwCommandHasIoPinInt')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandHasIoPinInt')

  if hasattr(lib, 'hwCommandHasIoPinFloat'):
    lib_func = getattr(lib, 'hwCommandHasIoPinFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiCommandRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwCommandHasIoPinFloat')

  if hasattr(lib, 'hwFeedbackGetFloat'):
    lib_func = getattr(lib, 'hwFeedbackGetFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetFloat')

  if hasattr(lib, 'hwFeedbackGetHighResAngle'):
    lib_func = getattr(lib, 'hwFeedbackGetHighResAngle')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_double),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetHighResAngle')

  if hasattr(lib, 'hwFeedbackGetVector3f'):
    lib_func = getattr(lib, 'hwFeedbackGetVector3f')
    lib_func.argtypes = [ctypes.POINTER(HebiVector3f),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetVector3f')

  if hasattr(lib, 'hwFeedbackGetQuaternionf'):
    lib_func = getattr(lib, 'hwFeedbackGetQuaternionf')
    lib_func.argtypes = [ctypes.POINTER(HebiQuaternionf),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetQuaternionf')

  if hasattr(lib, 'hwFeedbackGetUInt64'):
    lib_func = getattr(lib, 'hwFeedbackGetUInt64')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_uint64),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetUInt64')

  if hasattr(lib, 'hwFeedbackGetEnum'):
    lib_func = getattr(lib, 'hwFeedbackGetEnum')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetEnum')

  if hasattr(lib, 'hwFeedbackGetBool'):
    lib_func = getattr(lib, 'hwFeedbackGetBool')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetBool')

  if hasattr(lib, 'hwFeedbackGetNumberedFloat'):
    lib_func = getattr(lib, 'hwFeedbackGetNumberedFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetNumberedFloat')

  if hasattr(lib, 'hwFeedbackGetIoPin'):
    lib_func = getattr(lib, 'hwFeedbackGetIoPin')
    lib_func.argtypes = [ctypes.POINTER(HebiIoBankPinStruct),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetIoPin')

  if hasattr(lib, 'hwFeedbackGetIoPinInt'):
    lib_func = getattr(lib, 'hwFeedbackGetIoPinInt')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int64),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetIoPinInt')

  if hasattr(lib, 'hwFeedbackGetIoPinFloat'):
    lib_func = getattr(lib, 'hwFeedbackGetIoPinFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetIoPinFloat')

  if hasattr(lib, 'hwFeedbackGetLed'):
    lib_func = getattr(lib, 'hwFeedbackGetLed')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackGetLed')

  if hasattr(lib, 'hwFeedbackHasField'):
    lib_func = getattr(lib, 'hwFeedbackHasField')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackHasField')

  if hasattr(lib, 'hwFeedbackHasIoPinInt'):
    lib_func = getattr(lib, 'hwFeedbackHasIoPinInt')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackHasIoPinInt')

  if hasattr(lib, 'hwFeedbackHasIoPinFloat'):
    lib_func = getattr(lib, 'hwFeedbackHasIoPinFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiFeedbackRef),ctypes.c_uint32,ctypes.c_int,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwFeedbackHasIoPinFloat')

  if hasattr(lib, 'hwInfoGetFloat'):
    lib_func = getattr(lib, 'hwInfoGetFloat')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_float),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetFloat')

  if hasattr(lib, 'hwInfoGetHighResAngle'):
    lib_func = getattr(lib, 'hwInfoGetHighResAngle')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_double),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetHighResAngle')

  if hasattr(lib, 'hwInfoGetFlag'):
    lib_func = getattr(lib, 'hwInfoGetFlag')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetFlag')

  if hasattr(lib, 'hwInfoGetBool'):
    lib_func = getattr(lib, 'hwInfoGetBool')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetBool')

  if hasattr(lib, 'hwInfoGetEnum'):
    lib_func = getattr(lib, 'hwInfoGetEnum')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetEnum')

  if hasattr(lib, 'hwInfoGetLed'):
    lib_func = getattr(lib, 'hwInfoGetLed')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_int32),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoGetLed')

  if hasattr(lib, 'hwInfoHasField'):
    lib_func = getattr(lib, 'hwInfoHasField')
    lib_func.argtypes = [ctypes.POINTER(ctypes.c_bool),ctypes.POINTER(HebiInfoRef),ctypes.c_uint32,ctypes.c_uint32]
    lib_func.restype = None
  else:
    raise RuntimeWarning('Could not load function hwInfoHasField')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


from ctypes import byref, cast, c_ubyte, c_char_p, c_void_p, c_size_t, POINTER

from .group import GroupDelegate, Group
from .type_utils import (create_string_buffer_compat, decode_string_buffer, to_mac_address)
from .ffi.enums import StatusSuccess
from .ffi import ctypes_func_defs as api
from .ffi.wrappers import UnmanagedObject
from .ffi.ctypes_defs import HebiMacAddress


################################################################################
# Mac Address
################################################################################


class MacAddress(object):
  """
  A simple wrapper class for internal C-API HebiMacAddress objects to allow
  interfacing with API calls that use MAC addresses.
  """

  __slots__ = ['_obj']

  def __init__(self, a, b, c, d, e, f):
    obj = HebiMacAddress()
    obj.bytes_[0:6] = [a, b, c, d, e, f]

    self._obj = obj

  def __repr__(self):
    return self.__human_readable()

  def __str__(self):
    return self.__human_readable()

  def __getitem__(self, item):
    return self._obj.bytes_[item]

  def __human_readable(self):
    b0 = "%0.2X" % self._obj.bytes_[0]
    b1 = "%0.2X" % self._obj.bytes_[1]
    b2 = "%0.2X" % self._obj.bytes_[2]
    b3 = "%0.2X" % self._obj.bytes_[3]
    b4 = "%0.2X" % self._obj.bytes_[4]
    b5 = "%0.2X" % self._obj.bytes_[5]
    return '{0}:{1}:{2}:{3}:{4}:{5}'.format(b0, b1, b2, b3, b4, b5)

  @property
  def _as_parameter_(self):
    return self._obj

  @property
  def raw_bytes(self):
    """
    An unsigned byte buffer view of the object (ctypes c_ubyte array).
    Use this if you need a serialized format of this object, or if you
    are marshalling data to an external C API, etc.
    """
    return self._obj.bytes_


################################################################################
# Lookup Entries
################################################################################


class Entry(object):
  """
  Represents a HEBI module. This is used by the Lookup class.
  """

  __slots__ = ['_family', '_mac_address', '_name']

  def __init__(self, name, family, mac_address):
    self._name = name
    self._family = family
    self._mac_address = mac_address

  def __repr__(self):
    return self.__human_readable()

  def __str__(self):
    return self.__human_readable()

  def __human_readable(self):
    return 'Family: {} Name: {} Mac Address: {}'.format(self._family, self._name, self._mac_address)

  @property
  def name(self):
    """
    :return: The name of the module.
    :rtype:  str
    """
    return self._name

  @property
  def family(self):
    """
    :return: The family to which this module belongs.
    :rtype:  str
    """
    return self._family

  @property
  def mac_address(self):
    """
    :return: The immutable MAC address of the module.
    :rtype:  str
    """
    return self._mac_address


class EntryList(UnmanagedObject):
  """
  A list of module entries. This is used by the :class:`~hebi.Lookup` class
  and is returned by :attr:`~hebi.Lookup.entrylist`.
  """
  __slots__ = ['_elements', '_iterator', '_size']

  def __init__(self, internal):
    super(EntryList, self).__init__(internal, on_delete=api.hebiLookupEntryListRelease)
    bypass_debug_printing = True
    if bypass_debug_printing: #with bypass_debug_printing:
      # In debug mode, this tries to call repr(self) before everything is initialized,
      # which makes Python complain. So scope to bypass debug printing.
      # If debug mode is not enabled, this does not change any behavior.
      self._size = api.hebiLookupEntryListGetSize(self)
      elements = list()
      for i in range(self._size):
        elements.append(self.__get_entry(i))
      self._elements = elements
      self._iterator = iter(elements)

  def __iter__(self):
    return self

  def __length_hint__(self):
    return self._iterator.__length_hint__()

  def __next__(self):
    try:
      return next(self._iterator)
    except StopIteration:
      # PEP 479 forbids the implicit propagation of StopIteration
      raise StopIteration

  def __repr__(self):
    return self.__human_readable()

  def __str__(self):
    return str([str(entry) for entry in self._elements])

  def __human_readable(self):
    modules = list()
    for entry in self._elements:
      modules.append(entry)
    from .utils import lookup_table_string
    return lookup_table_string(modules)

  def __get_entry(self, index):
    required_size = c_size_t(0)
    if (api.hebiLookupEntryListGetName(self, index, c_char_p(None), byref(required_size)) != StatusSuccess):
      return None

    c_buffer = create_string_buffer_compat(required_size.value)
    if (api.hebiLookupEntryListGetName(self, index, c_buffer, byref(required_size)) != StatusSuccess):
      return None

    name = decode_string_buffer(c_buffer, 'utf-8')
    if (api.hebiLookupEntryListGetFamily(self._internal, index, c_char_p(None), byref(required_size)) != StatusSuccess):
      return None

    c_buffer = create_string_buffer_compat(required_size.value)
    if (api.hebiLookupEntryListGetFamily(self._internal, index, c_buffer, byref(required_size)) != StatusSuccess):
      return None

    family = decode_string_buffer(c_buffer, 'utf-8')
    tmp_buffer = HebiMacAddress()
    if (api.hebiLookupEntryListGetMacAddress(self._internal, index, byref(tmp_buffer)) != StatusSuccess):
      return None

    return Entry(name, family, MacAddress(*tmp_buffer.bytes_))

  def __getitem__(self, index):
    return self.__get_entry(index)


################################################################################
# Lookup and delegates
################################################################################


class LookupDelegate(UnmanagedObject):
  """
  Delegate for Lookup
  """

  __slots__ = []

  __singleton = None
  __singleton_lock = None

  @staticmethod
  def get_singleton():
    LookupDelegate.__singleton_lock.acquire()
    if LookupDelegate.__singleton is None:
      LookupDelegate.__singleton = LookupDelegate()
    LookupDelegate.__singleton_lock.release()
    return LookupDelegate.__singleton


  def __parse_to(self, timeout_ms):
    if timeout_ms is None:
      # FIXME: don't import here. We have to ATM because there would be cyclic dependencies otherwise
      from .. import Lookup
      return Lookup.DEFAULT_TIMEOUT_MS
    else:
      try:
        return int(timeout_ms)
      except:
        raise ValueError('timeout_ms must be a number')

  def __init__(self):
    super(LookupDelegate, self).__init__(api.hebiLookupCreate(None, 0), on_delete=api.hebiLookupRelease)

  @property
  def entrylist(self):
    list = api.hebiCreateLookupEntryList(self)
    if list:
      return EntryList(list)
    else:
      return None

  @property
  def lookup_frequency(self):
    return api.hebiLookupGetLookupFrequencyHz(self)

  @lookup_frequency.setter
  def lookup_frequency(self, freq):
    api.hebiLookupSetLookupFrequencyHz(self, freq)

  def get_group_from_names(self, families, names, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    families_length = len(families)
    names_length = len(names)

    families_buffer = (c_char_p * families_length)()
    families_buffer_list = [ None ] * families_length
    names_buffer = (c_char_p * names_length)()
    names_buffer_list = [ None ] * names_length

    for i, family in enumerate(families):
      family_length = len(family)+1
      families_buffer_list[i] = create_string_buffer_compat(family, family_length)

    for i, name in enumerate(names):
      name_length = len(name)+1
      names_buffer_list[i] = create_string_buffer_compat(name, name_length)

    for i in range(families_length):
      families_buffer[i] = cast(families_buffer_list[i], c_char_p)

    for i in range(names_length):
      names_buffer[i] = cast(names_buffer_list[i], c_char_p)

    c_char_pp = POINTER(c_char_p)
    group = c_void_p(api.hebiGroupCreateFromNames(self,
                                              cast(byref(families_buffer), c_char_pp),
                                              families_length,
                                              cast(byref(names_buffer), c_char_pp),
                                              names_length,
                                              timeout_ms))

    if group:
      return Group(GroupDelegate(group))
    return None

  def get_group_from_macs(self, addresses, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    addresses_length = len(addresses)
    addresses_list = [None] * addresses_length

    for i, address in enumerate(addresses):
      addresses_list[i] = to_mac_address(address)

    addresses_list_c = (c_void_p * addresses_length)()
    for i in range(addresses_length):
      addresses_list_c[i] = cast(addresses_list[i].raw_bytes, c_void_p)

    group = c_void_p(api.hebiGroupCreateFromMacs(self, byref(addresses_list_c),
                                             addresses_length, timeout_ms))

    if group:
      return Group(GroupDelegate(group))
    return None

  def get_group_from_family(self, family, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    family_buffer = create_string_buffer_compat(family, len(family)+1)
    group = c_void_p(api.hebiGroupCreateFromFamily(self, family_buffer, timeout_ms))

    if (group):
      return Group(GroupDelegate(group))
    return None

  def get_connected_group_from_name(self, family, name, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    family_buffer = create_string_buffer_compat(family, len(family)+1)
    name_buffer = create_string_buffer_compat(name, len(name)+1)
    group = c_void_p(api.hebiGroupCreateConnectedFromName(self, family_buffer,
                                                      name_buffer, timeout_ms))

    if group:
      return Group(GroupDelegate(group))
    return None

  def get_connected_group_from_mac(self, address, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    mac_address = to_mac_address(address)
    group = c_void_p(api.hebiGroupCreateConnectedFromMac(self, mac_address.raw_bytes,
                                                     timeout_ms))

    if group:
      return Group(GroupDelegate(group))
    return None


from threading import Lock
LookupDelegate._LookupDelegate__singleton_lock = Lock()
del Lock



def parse_version(s):
  split = s.split('.')
  if len(split) > 3:
    raise ValueError("Period characters are not allowed in the suffix portion of the version string `{0}`".format(s))
  elif len(split) < 3:
    raise ValueError("Malformed version string `{0}`".format(s))

  import re
  r = re.compile(r"^(\d+)([a-zA-Z_0-9-]+)?$")
  patch_str = split[2]
  patch_parts = r.split(patch_str)
  parts = []
  for entry in patch_parts:
    if entry != '':
      parts.append(entry)

  if len(parts) > 2 or len(parts) == 0:
    raise ValueError(s)
  elif len(parts) == 1:
    return Version(int(split[0]), int(split[1]), int(parts[0]))
  return Version(int(split[0]), int(split[1]), int(parts[0]), parts[1])


class Version(object):
  def __init__(self, maj, min, rev, suffix=None):
    """
    Suffix string must only contain the following:
      * Underscore
      * Hyphen
      * alphabetical character (lower or uppercase) as determined by `[a-zA-Z]
      * numbers as determined by `[0-9]` or `\d`
    
    Note that periods (`.`) are not allowed in the suffix.
    """
    self._major_version = maj
    self._minor_version = min
    self._patch_version = rev
    self._suffix = suffix or ''

  def __str__(self):
    return '{0}.{1}.{2}{3}'.format(self.major_version, self.minor_version, self.patch_version, self.suffix)

  def __repr__(self):
    return 'Version(major: {0}, minor: {1}, patch: {2}, suffix: {3})'.format(self.major_version, self.minor_version, self.patch_version, self.suffix)

  def __eq__(self, o):
    if isinstance(o, Version):
      thiz_suffix = self._suffix or ''
      o_suffix = o._suffix or ''
      return o.major_version == self.major_version and \
      o.minor_version == self.minor_version and \
      o.patch_version == self.patch_version and \
      thiz_suffix == o_suffix
    else:
      return self == parse_version(o)

  def __ne__(self, o):
    if isinstance(o, Version):
      thiz_suffix = self._suffix or ''
      o_suffix = o._suffix or ''
      return o.major_version != self.major_version or \
      o.minor_version != self.minor_version or \
      o.patch_version != self.patch_version or \
      thiz_suffix != o_suffix
    else:
      return self != parse_version(o)

  def __gt__(self, o):
    if isinstance(o, Version):
      if self.major_version > o.major_version:
        return True
      elif self.major_version < o.major_version:
        return False
      elif self.minor_version > o.minor_version:
        return True
      elif self.minor_version < o.minor_version:
        return False
      elif self.patch_version > o.patch_version:
        return True
      elif self.patch_version < o.patch_version:
        return False
      thiz_suffix = self._suffix or ''
      o_suffix = o._suffix or ''
      return thiz_suffix > o_suffix
    else:
      return self > parse_version(o)

  def __lt__(self, o):
    if isinstance(o, Version):
      if self.major_version < o.major_version:
        return True
      elif self.major_version > o.major_version:
        return False
      elif self.minor_version < o.minor_version:
        return True
      elif self.minor_version > o.minor_version:
        return False
      elif self.patch_version < o.patch_version:
        return True
      elif self.patch_version > o.patch_version:
        return False
      thiz_suffix = self._suffix or ''
      o_suffix = o._suffix or ''
      return thiz_suffix < o_suffix
    else:
      return self < parse_version(o)

  @property
  def major_version(self):
    return self._major_version

  @property
  def minor_version(self):
    return self._minor_version

  @property
  def patch_version(self):
    return self._patch_version

  @property
  def suffix(self):
    return self._suffix
  


def py_version():
  return parse_version("2.1.0")


def min_c_api_version():
  return parse_version("2.0.0")


def loaded_c_api_version():
  from ._internal.ffi.loader import _handle
  return _handle.version


if __name__ == "__main__":
  import argparse
  def __disp(txt, vfunc):
    print('{0}: {1}'.format(txt, vfunc()))

  parser = argparse.ArgumentParser()
  parser.add_argument("--min-c-api", help="Show minimum required C API version",
                      default=False, action="store_true")
  parser.add_argument("--py-api", help="Show the version of hebi-py",
                      default=False, action="store_true")
  args = parser.parse_args()
  if args.min_c_api:
    __disp('Minimum C API Version', min_c_api_version)
  if args.py_api:
    __disp('hebi-py Version', py_version)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# ------------------------------------------------------------------------------

from ctypes import c_double, byref
import numpy as np
from .ffi.enums import StatusSuccess
from .ffi.ctypes_func_defs import hebiTrajectoryGetState


class Trajectory(object):
  """
  Represents a smooth trajectory through a set of waypoints.
  """

  __slots__ = ['_trajectories', '_waypoint_times', '_number_of_waypoints', '_number_of_joints', '_start_time', '_end_time']

  def __init__(self, trajectories, waypoint_times, num_waypoints):
    self._trajectories = trajectories
    self._waypoint_times = waypoint_times
    self._number_of_waypoints = num_waypoints
    self._number_of_joints = len(trajectories)
    self._start_time = waypoint_times[0]
    self._end_time = waypoint_times[-1]

  @property
  def number_of_waypoints(self):
    """
    The number of waypoints in this trajectory.

    :return: number of waypoints
    :rtype:  int
    """
    return self._number_of_waypoints

  @property
  def number_of_joints(self):
    """
    The number of joints in this trajectory.

    :return: the number of joints
    :rtype:  int
    """
    return len(self._trajectories)

  @property
  def start_time(self):
    """
    The time (in seconds) at which the trajectory starts.

    :return: the start time
    :rtype:  float
    """
    return self._start_time

  @property
  def end_time(self):
    """
    The time (in seconds) at which the trajectory ends.

    :return: the end time
    :rtype:  float
    """
    return self._end_time

  @property
  def duration(self):
    """
    The time (in seconds) between the start and end of this trajectory.

    :return: the duration
    :rtype:  float
    """
    return self._end_time - self._start_time

  @property
  def waypoint_times(self):
    """

    :return: The input time (in seconds) for each waypoint
    :rtype:  numpy.ndarray
    """
    return self._waypoint_times

  def get_state(self, time):
    """
    Returns the position, velocity, and acceleration for a given point
    in time along the trajectory.

    :param time: the time, in seconds
    :type time:  int, float

    :return: a triplet containing the position, velocity, and acceleration
             at the given point in time.
    :rtype:  numpy.ndarray, numpy.ndarray, numpy.ndarray
    """
    position_val = c_double(0.0)
    velocity_val = c_double(0.0)
    acceleration_val = c_double(0.0)

    position = np.zeros(self.number_of_joints, np.float64)
    velocity = np.zeros(self.number_of_joints, np.float64)
    acceleration = np.zeros(self.number_of_joints, np.float64)

    for i, trajectory in enumerate(self._trajectories):
      passed_run =\
        hebiTrajectoryGetState(trajectory, time,
                               byref(position_val),
                               byref(velocity_val),
                               byref(acceleration_val)) == StatusSuccess

      position[i] = position_val.value
      velocity[i] = velocity_val.value
      acceleration[i] = acceleration_val.value

    return position, velocity, acceleration



from os.path import join, dirname, realpath
from platform import architecture, machine

from .. import debug


class HEBICoreLibrary(object):
  """
  Loads the C API functions. Do not use.
  """
  __slots__ = ['__core_api_version', '__core_lib', '__core_location', '__wrapper_lib', '__wrapper_location']

  def __init__(self, core_lib, core_location, wrapper_lib, wrapper_locaton):
    self.__core_lib = core_lib
    self.__core_location = core_location
    self.__wrapper_lib = wrapper_lib
    self.__wrapper_location = wrapper_locaton

    from .ctypes_loader import populate_hebi_funcs, populate_hebi_wrapper_funcs
    populate_hebi_funcs(core_lib)
    populate_hebi_wrapper_funcs(wrapper_lib)

    from .ctypes_func_defs import set_api_handle, set_api_wrapper_handle
    set_api_handle(core_lib)
    set_api_wrapper_handle(wrapper_lib)

    from .. import Version, min_c_api_version
    self.__core_api_version = Version(-1, -1, -1)

    # Make sure to check for a compatible version of the C API first
    if hasattr(core_lib, 'hebiGetLibraryVersion'):
      c_api_version = min_c_api_version()
      c_min_major = c_api_version.major_version
      c_min_minor = c_api_version.minor_version
      c_min_patch = c_api_version.patch_version

      from ctypes import byref, c_int, c_int32, POINTER
      from .. import debug

      major_version = c_int32(0)
      minor_version = c_int32(0)
      patch_version = c_int32(0)
      core_lib.hebiGetLibraryVersion(byref(major_version), byref(minor_version), byref(patch_version))

      req_str = '{}.{}.{}'.format(c_min_major, c_min_minor, c_min_patch)
      cur_str = '{}.{}.{}'.format(major_version.value, minor_version.value, patch_version.value)
      debug.debug_log('hebiGetLibraryVersion() ==> {}'.format(cur_str))

      if major_version.value != 2:
        # Refuse to load anything which is not a 2.x.x binary
        raise RuntimeError("C API library must be a 2.x.x release (loaded version {})".format(cur_str))

      if cur_str < req_str:
        print('Warning: loaded C library may be incompatible, as it is an outdated library. ' +\
          'Loaded library version is {}, required minimum is {}'.format(cur_str, req_str))

      self.__core_api_version._major_version = major_version.value
      self.__core_api_version._minor_version = minor_version.value
      self.__core_api_version._patch_version = patch_version.value


    from .ctypes_defs import HebiCommandMetadata, HebiFeedbackMetadata, HebiInfoMetadata

    command_metadata = HebiCommandMetadata()
    feedback_metadata = HebiFeedbackMetadata()
    info_metadata = HebiInfoMetadata()

    core_lib.hebiCommandGetMetadata(byref(command_metadata))
    core_lib.hebiFeedbackGetMetadata(byref(feedback_metadata))
    core_lib.hebiInfoGetMetadata(byref(info_metadata))
    wrapper_lib.hwInitialize(byref(command_metadata), byref(feedback_metadata), byref(info_metadata))

  def __del__(self):
    self.__core_lib.hebiCleanup()

  @property
  def core_lib_location(self) -> str:
    return self.__core_location

  @property
  def wrapper_lib_location(self) -> str:
    return self.__wrapper_location

  @property
  def version(self):
    from .. import Version
    maj_v = self.__core_api_version.major_version
    min_v = self.__core_api_version.minor_version
    pat_v = self.__core_api_version.patch_version
    return Version(maj_v, min_v, pat_v)


class SharedLibraryLoader(object):
  def __init__(self, library):
    self._candidates = list()
    self._library = library

  def add_candidate(self, loc):
    self._candidates.append(loc)

  @property
  def candidates(self) -> list:
    return self._candidates

  def try_load_library(self):
    """
    Goes through all candidate libraries in priority order.

    :return: a tuple (pair): `ctypes` object representing the library, along with the file path of the library 
    """
    candidates = self.candidates
    candidate_count = len(candidates)
    debug.debug_log("Loading library {}.".format(self._library))
    debug.debug_log("Loader(library={}): {} candidate binaries:".format(self._library, candidate_count))
    debug.debug_log("Loader(library={}): Candidate binaries, in order of priority:".format(self._library))

    # Loop twice so all candidate libraries can be printed to debug stream in case of debugging enabled
    for i, candidate in enumerate(candidates):
      debug.debug_log("Candidate {}: {}".format(i+1, candidate))

    from ctypes import cdll
    for candidate in candidates:
      try:
        lib = cdll.LoadLibrary(candidate)
        debug.debug_log('Successfully loaded library at {}'.format(candidate))
        return lib, candidate
      except Exception as e:
        debug.debug_log("Attempting to load library {} raised exception:\n{}".format(candidate, e))

    debug.warn_log("Unable to load (library={}).\nCandidate libraries attempted (in order):".format(self._library))
    for i, candidate in enumerate(candidates):
      debug.warn_log("Candidate {}: {}".format(i+1, candidate))
    return None, None


def _get_library_load_candidates(library, readable_name, env_var=None):
  ret = list()
  from os import environ
  import sys

  # Top priority: environment variables (if set)
  if env_var is not None:
    environ_val = environ.get(env_var)
    if environ_val is not None:
      debug.debug_log("{} library candidate '{}' from {} environment variable".format(readable_name, environ_val, env_var))
      ret.append(environ_val)

  # Lower priority: load from installed package
  lib_base_path = join(join(dirname(realpath(__file__)), '..', '..'), 'lib')
  
  from hebi import version
  c_api_version = version.min_c_api_version()
  maj_ver = c_api_version.major_version
  min_ver = c_api_version.minor_version

  if sys.platform.startswith('linux'):
    _find_linux_candidates(ret, lib_base_path, maj_ver, min_ver, library)
  elif sys.platform == 'darwin':
    _find_mac_candidates(ret, lib_base_path, maj_ver, library)
  elif sys.platform == 'win32':
    _find_win_candidates(ret, lib_base_path, library)

  return ret


def _load_shared_library(name, readable_name, env_var=None):
  loader = SharedLibraryLoader(name)

  for entry in _get_library_load_candidates(name, readable_name, env_var):
    loader.add_candidate(entry)

  loaded_c_lib, loaded_loc = loader.try_load_library()
  if loaded_c_lib is None:
    raise RuntimeError('{} library not found'.format(readable_name))

  return loaded_c_lib, loaded_loc


_load_core_shared_library = lambda: _load_shared_library('hebi', 'HEBI Core', 'HEBI_C_LIB')
_load_wrapper_shared_library = lambda: _load_shared_library('hebiWrapper', 'HEBI Core wrapper', 'HEBI_WRAPPER_LIB')


def _find_linux_candidates(output, lib_base_path, maj_ver, min_ver, library):
  import re
  cpu = machine()
  py_exec_arch = architecture()[0]
  lib_str = 'lib{}.so'.format(library)

  if cpu == 'x86_64' and ('64' in py_exec_arch):
    # 64 bit x86 CPU with 64 bit python
    lib_path = join(lib_base_path, 'linux_x86_64', lib_str)

  elif ((re.match('i[3-6]86', cpu) is not None)
        or (cpu == 'x86_64') and ('32' in py_exec_arch)):
    raise RuntimeError('i686 is no longer supported. If you are on a 64 bit kernel, install and run an x86_64 instance of Python.')

  elif (re.match('arm.*', cpu) is not None) and ('32' in py_exec_arch):
    # 32 bit armhf with 32 bit python
    lib_path = join(lib_base_path, 'linux_armhf', lib_str)

  elif ((re.match('arm.*', cpu) is not None)
        or 'aarch64' in cpu and ('64' in py_exec_arch)):
    lib_path = join(lib_base_path, 'linux_aarch64', lib_str)
  else:
    raise RuntimeError('Unknown architecture {0}'.format(cpu))

  output.append('{}.{}.{}'.format(lib_path, maj_ver, min_ver))
  output.append('{}.{}'.format(lib_path, maj_ver))
  output.append(lib_path)


def _find_mac_candidates(output, lib_base_path, maj_ver, library):
  output.append(join(lib_base_path, 'osx_amd64', 'lib{}.{}.dylib'.format(library, maj_ver)))
  output.append(join(lib_base_path, 'osx_amd64', 'lib{}.dylib'.format(library)))


def _find_win_candidates(output, lib_base_path, library):
  cpu = machine()
  py_exec_arch = architecture()[0]

  if cpu == 'AMD64' or cpu == 'x86':
    # Windows doesn't like to make it easy to detect which architecture the process is running in (x86 vs x64)
    # You can use `ctypes` to detect this, but this is a more terse way.
    output.append(join(lib_base_path, 'win_x64', '{}.dll'.format(library)))
    output.append(join(lib_base_path, 'win_x86', '{}.dll'.format(library)))
  elif cpu == 'ARM':
    # XXX Not yet supported :(
    # 32 bit ARM on Windows
    raise RuntimeError('ARM is not yet supported on Windows')
  elif cpu == 'ARM64':
    # XXX Not yet supported :(
    # 64 bit ARM on Windows
    raise RuntimeError('ARM64 is not yet supported on Windows')
  else:
    raise RuntimeError('Unknown architecture {}'.format(cpu))


def _init_libraries():
  core_lib, core_loc = _load_core_shared_library()
  wrapper_lib, wrapper_loc = _load_wrapper_shared_library()
  
  return HEBICoreLibrary(core_lib, core_loc, wrapper_lib, wrapper_loc)


# Load library on import
_handle = _init_libraries()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# ------------------------------------------------------------------------------

from ._internal import log_file as _log_file
from ._internal import group as _group

from os.path import isfile as _isfile

try:
  import imp
  imp.find_module('matplotlib')
  _found_matplotlib = True


  def _field_is_command(field_name):
    return field_name.snake_case in ['position', 'velocity', 'effort']


except ImportError:
  _found_matplotlib = False

if not _found_matplotlib:
  print('matplotlib not found - hebi.util.plot_logs and hebi.util.plot_trajectory will not work.')


def create_imitation_group(size):
  """
  Create an imitation group of the provided size.
  The imitation group returned from this function provides the exact same
  interface as a group created from the :class:`Lookup` class.

  However, there are a few subtle differences between the imitation group and
  group returned from a lookup operation. See :ref:`imitation-group-contrast` section
  for more information.

  :param size: The number of modules in the imitation group
  :type size:  int
  
  :return: The imitation group. This will never be ``None``
  :rtype:  Group

  :raises ValueError: If size is less than 1
  """
  return _group.create_imitation_group(size)


def load_log(file):
  """
  Opens an existing log file.

  :param file: the path to an existing log file
  :type file:  str, unicode

  :return: The log file. This function will never return ``None``
  :rtype:  LogFile
  
  :raises TypeError: If file is an invalid type
  :raises IOError: If the file does not exist or is not a valid log file
  """
  try:
    f_exists = _isfile(file)
  except TypeError as t:
    raise TypeError('Invalid type for file. Caught TypeError with message: {}'.format(t.args))

  if not f_exists:
    raise IOError('file {} does not exist'.format(file))

  from ._internal.ffi.ctypes_func_defs import hebiLogFileOpen
  log_file = hebiLogFileOpen(file.encode('utf-8'))
  if log_file is None:
    raise IOError('file {} is not a valid log file'.format(file))

  return _log_file.LogFile(log_file)


def plot_logs(logs, fbk_field, figure_spec=None, modules=None):
  """
  Nicely formatted plotting of HEBI logs.

  :param logs: The log(s) (or file path to a log) to plot
  :type logs: list, LogFile, str

  :param fbk_field: Feedback field to plot
  :type fbk_field:  str

  :param figure_spec: The figure number you would like to use for plots. If multiple log files
                      are plotted, then the subsequent figure numbers increment by 1 starting
                      at `figure_spec`. If unspecified, a new figure will be created to avoid
                      overwriting previous figures.
  :type figure_spec:  int

  :param modules: Optionally select which modules to plot
  :type modules: NoneType, list
  """
  from matplotlib import pyplot as plt
  import numpy as np
  if isinstance(logs, _log_file.LogFile):
    logs = [logs]
  elif isinstance(logs, str):
    logs = [logs]
  elif not isinstance(logs, list):
    raise TypeError('Parameter logs was of unexpected type {}'.format(type(logs).__name__))

  # Convert any `str` to LogFile
  for i, entry in enumerate(logs[:]):
    if isinstance(entry, str):
      logs[i] = load_log(entry)

  from ._internal.field_bindings import get_field_info
  feedback_info = get_field_info(fbk_field)
  command_plot = _field_is_command(feedback_info)
  num_logs = len(logs)

  for i in range(num_logs):
    log = logs[i]
    if modules is None:
      plot_mask = [int(j) for j in range(log.number_of_modules)]
    else:
      plot_mask = modules

    if figure_spec is None:
      fig = plt.figure()
    else:
      curr_fig = figure_spec + i
      fig = plt.figure(curr_fig, clear=True)

    if command_plot:
      ax = plt.subplot(2, 1, 1)
      ax2 = plt.subplot(2, 1, 2)
      plt.sca(ax)
    else:
      ax = plt.axes()

    num_plots = len(plot_mask)

    x_module_series = list()
    y_module_series = list()
    cmd_module_series = list()
    diff_module_series = list()

    for _ in range(num_plots):
      x_module_series.append(list())
      y_module_series.append(list())
      cmd_module_series.append(list())
      diff_module_series.append(list())

    get_y = lambda entry: feedback_info.get_field(entry)[plot_mask]
    x_lim_max = 0
    x_lim_min = None

    if command_plot:
      cmd_info = get_field_info(feedback_info.snake_case + "_command")
      get_cmd = lambda entry: cmd_info.get_field(entry)[plot_mask]

      last_entry = None

      for entry in log.feedback_iterate:
        last_entry = entry
        entry_time = entry.time[plot_mask]
        entry_y    = get_y(entry)
        entry_cmd  = get_cmd(entry)

        if x_lim_min is None:
          x_lim_min = entry_time.min()

        for j in range(num_plots):
          x = entry_time[j]
          y = entry_y[j]
          cmd = entry_cmd[j]
          diff = y - cmd

          x_module_series[j].append(x)
          y_module_series[j].append(y)
          cmd_module_series[j].append(cmd)
          diff_module_series[j].append(diff)
      x_lim_max = last_entry.time[plot_mask].max()
    else:
      last_entry = None
      for entry in log.feedback_iterate:
        last_entry = entry
        entry_time = entry.time[plot_mask]
        entry_y = get_y(entry)

        if x_lim_min is None:
          x_lim_min = entry_time.min()

        for j in range(num_plots):
          x = entry_time[j]
          y = entry_y[j]
          x_module_series[j].append(x)
          y_module_series[j].append(y)
      x_lim_max = last_entry.time[plot_mask].max()

    y_label_unit_str = ' ({0})'.format(feedback_info.units)

    for j in range(num_plots):
      plt.plot(x_module_series[j], y_module_series[j])

    plt.xlabel('time (sec)')
    plt.ylabel(feedback_info.pascal_case + y_label_unit_str)
    plt.title('{} - Log {} of {}'.format(feedback_info.pascal_case, i + 1, num_logs))
    plt.xlim(x_lim_min, x_lim_max)
    plt.grid(True)

    if command_plot:
      plt.sca(ax2)
      plt.xlabel('time (sec)')
      plt.ylabel('error' + y_label_unit_str)
      plt.title('{} error'.format(feedback_info.pascal_case))
      plt.xlim(x_lim_min, x_lim_max)
      plt.grid(True)

      for j in range(num_plots):
        plt.sca(ax)
        ax.ColorOrderIndex = 1
        plt.plot(x_module_series[j], cmd_module_series[j], '--')

        plt.sca(ax2)
        plt.plot(x_module_series[j], diff_module_series[j])

    plt.legend([str(mask) for mask in plot_mask])

  plt.show()


def plot_trajectory(trajectory, dt=0.01, figure_spec=None, legend=None):
  """
  Visualizes position, velocity, and acceleration of a trajectory.

  :param trajectory:
  :type trajectory:  hebi.trajectory.Trajectory

  :param dt: Delta between points in trajectory to plot
  :type dt:  int, float

  :param figure_spec: The figure number or figure handle that should be used for plotting.
                      If a figure with the specified number exists it will be overwritten.
                      If left unspecified, a new figure will automatically be generated.
  :type figure_spec:  int, str

  :param legend: String of the text that gets displayed as the legend.
                 By default it shows the joint number.
  :type legend:  str
  """
  from matplotlib import pyplot as plt
  import numpy as np

  if figure_spec is None:
    fig = plt.figure()
  elif isinstance(figure_spec, (int, str)):
    fig = plt.figure(num=figure_spec)
  else:
    raise TypeError('Parameter figure_spec was of unexpected type {}.'.format(type(figure_spec).__name__))

  if legend is None:
    # Joint number::: TODO
    legend = ''

  duration = trajectory.duration
  linear_times = np.arange(0.0, duration, dt)
  actual_times = trajectory.waypoint_times

  linear_pos = [0.0] * len(linear_times)
  linear_vel = [0.0] * len(linear_times)
  linear_acc = [0.0] * len(linear_times)
  actual_pos = [0.0] * len(actual_times)
  actual_vel = [0.0] * len(actual_times)
  actual_acc = [0.0] * len(actual_times)

  for i, t in enumerate(linear_times):
    p_t, v_t, a_t = trajectory.get_state(t)
    linear_pos[i] = p_t
    linear_vel[i] = v_t
    linear_acc[i] = a_t

  for i, t in enumerate(actual_times):
    p_t, v_t, a_t = trajectory.get_state(t)
    actual_pos[i] = p_t
    actual_vel[i] = v_t
    actual_acc[i] = a_t

  # NOTE:
  #   time          == linear_times
  #   waypointTime  == actual_times

  ax = plt.subplot(3, 1, 1)
  plt.plot(linear_times, linear_pos)
  ax.ColorOrderIndex = 1

  plt.plot(actual_times, actual_pos, marker='o')
  plt.title('Trajectory Profile')
  plt.ylabel('position (rad)')
  plt.xlabel('time (sec)')
  plt.grid(True)

  plt.legend(legend)

  ax = plt.subplot(3, 1, 2)
  plt.plot(linear_times, linear_vel)
  ax.ColorOrderIndex = 1
  plt.plot(actual_times, actual_vel, marker='o')
  plt.ylabel('velocity (rad/sec)')
  plt.xlabel('time (sec)')
  plt.grid(True)

  ax = plt.subplot(3, 1, 3)
  plt.plot(linear_times, linear_acc)
  ax.ColorOrderIndex = 1
  plt.plot(actual_times, actual_acc, marker='o')
  plt.ylabel('acceleration (rad/sec^2)')
  plt.xlabel('time (sec)')
  plt.grid(True)

  plt.show()


def clear_all_groups():
  """
  Clear all groups currently allocated by the API.
  This is useful to clear up resources when running in an environment such as IPython.
  """
  from ._internal.group import GroupDelegate
  GroupDelegate.destroy_all_instances()
